from lazy_rpc.backend.base import Sender, Receiver
from lazy_rpc.utils import pack, unpack, build_uniq_name
from lazy_rpc.decolators import FUNCTIONS

from kafka import KafkaConsumer, KafkaProducer
from multiprocessing import Process

class Kafka(Sender, Receiver):

  def __init__(self, servers=[]):
    self.servers = servers

  def setup_sender(self):
    self.producer = KafkaProducer(bootstrap_servers=self.servers)

  def setup_receiver(self):
    self.consumers = []

  def send(self, fn, *args, **kwargs):
    uniq_name = build_uniq_name(fn)
    value = pack(fn, *args, **kwargs)
    future = self.producer.send(uniq_name, value)
    result = future.get(timeout=60)

  def receive(self):
    plist = []
    for topic in FUNCTIONS.keys():
      fn = self.__build_process_fn(topic)
      p = Process(target=fn)
      plist.append(p)
      p.start()

    for p in plist:
      p.join()


  def __build_process_fn(self, topic):
    consumer = KafkaConsumer(topic, bootstrap_servers=self.servers)
    func = FUNCTIONS[topic]
    self.consumers.append(consumer)
    def __fn():
      for msg in consumer:
        Process(target=__unpack, args=[msg.value]).start()

    def __unpack(msg):
      value = unpack(msg)
      args = value['args']
      kwargs = value['kwargs']
      func(*args, **kwargs)

    return __fn





from lazy_rpc import remote
from lazy_rpc.utils import filtered_lazy_rpc_keyword, build_uniq_name

FUNCTIONS = {}

def lazy(fn):
  def proxy(*args, **kwargs):
    is_server = kwargs.get('l__sender', None) is None

    if is_server:
      kwargs = filtered_lazy_rpc_keyword(kwargs)
      fn(*args, **kwargs)
    else:
      remote.call(fn, *args, **kwargs)

  FUNCTIONS[build_uniq_name(fn)] = proxy

  return proxy




import setuptools

from lazy_rpc.version import VERSION

def _requires_from_file(filename):
  with open(filename, 'r') as f:
    ret = f.read().splitlines()
    if ret[0].startswith('-i http'):
      del ret[0]
    
    return ret

if __name__ == '__main__':

  extras_require = {
    'dev': _requires_from_file('requirements-dev.txt'),
  }
  
  with open('README.md', 'r') as f:
    long_description = f.read()


  setuptools.setup(
    name='lazy-rpc',
    version=VERSION,
    description='rpc module',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/m0cchi/lazy-rpc',
    packages=setuptools.find_packages(exclude=[
      'test',
      'test.*',
    ]),
    install_requires=_requires_from_file('requirements.txt'),
    extras_require=extras_require,

    license='MIT',
    python_requires='>=3.0',
    classifiers=[
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3',
      'Operating System :: OS Independent',
    ],

    author='m0cchi',
    author_email='m0cchi@protonmail.com',
  )



import pickle

def filtered_lazy_rpc_keyword(kwargs):
  return {k:v for k,v in kwargs.items() if is_not_lazy_rpc_keyword(k)}

def is_not_lazy_rpc_keyword(kw :str) -> bool:
  return kw is None or not kw.startswith('l__')

def build_uniq_name(fn):
  uniq_name = '{}--{}'.format(fn.__module__,
                              fn.__qualname__)
  return uniq_name

def pack(fn, *args, **kwargs) -> bytes:
    value = {
      'args': args,
      'kwargs': kwargs,
    }
    return pickle.dumps(value)

def unpack(value):
  return pickle.loads(value)



from lazy_rpc.utils import is_not_lazy_rpc_keyword

def call(fn, *args, **kwargs):
  code = fn.__code__
  len_kwargs = len(list(filter(is_not_lazy_rpc_keyword, kwargs.keys())))
  
  if code.co_argcount != len(args) or code.co_kwonlyargcount != len_kwargs:
    raise TypeError('type error')

  sender = kwargs['l__sender']
  del kwargs['l__sender']
  sender.send(fn, *args, **kwargs)




from abc import abstractmethod

class Sender:

  @abstractmethod
  def setup_sender(self):
    pass

  @abstractmethod
  def send(self):
    pass

class Receiver:

  @abstractmethod
  def setup_receiver(self):
    pass

  @abstractmethod
  def receive(self):
    pass

